import 'dart:math';
import 'dart:ui';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter/gestures.dart';

import 'package:yuro_core/yuro_core.dart';
import 'package:yuro_instance/yuro_instance.dart';

import 'core/binding.dart';
import 'core/middleware.dart';
import 'core/transition.dart';
import 'yuro_page.dart';
import 'yuro_route_ext.dart';

class YuroPageRoute<T> extends PageRoute<T> {
  YuroPageRoute({
    this.routeName,
    this.build,
    this.parameter,
    this.binding,
    this.middlewares,
    this.alignment,
    this.curve,
    this.transition,
    this.customTransition,
    this.transitionDuration = const Duration(milliseconds: 300),
    this.popGesture,
    this.opaque = true,
    this.barrierColor,
    this.barrierLabel,
    this.barrierDismissible = false,
    this.maintainState = true,
    RouteSettings settings,
    bool fullscreenDialog = false,
  })  : assert(opaque != null),
        assert(barrierDismissible != null),
        assert(maintainState != null),
        assert(fullscreenDialog != null),
        reference = '$routeName: ${build.hashCode}',
        super(settings: settings, fullscreenDialog: fullscreenDialog);

  final String routeName;
  final YuroPageBuild build;
  final Map<String, String> parameter;

  final Binding binding;
  final List<YuroMiddleware> middlewares;

  final Alignment alignment;
  final Curve curve;
  final Transition transition;
  final CustomTransition customTransition;

  final bool popGesture;

  final String reference;

  @override
  final bool opaque;

  @override
  final Duration transitionDuration;

  @override
  final Color barrierColor;

  @override
  final String barrierLabel;

  @override
  final bool barrierDismissible;

  @override
  final bool maintainState;

  @override
  Widget buildPage(BuildContext context, Animation<double> animation, Animation<double> secondaryAnimation) {
    Yuro.reference = reference;
    // runBindingStart ==> runPageBuildStart ==> runPageBuildEnd
    final middlewareRunner = MiddlewareRunner(middlewares);

    final waitBinding = middlewareRunner.runBindingStart(binding);
    waitBinding?.dependencies();

    final pageBuildEnd = middlewareRunner.runPageBuildStart(build);

    return middlewareRunner.runPageBuildEnd(pageBuildEnd());
  }

  @override
  void dispose() {
    super.dispose();
    if (Yuro.smartManagement != SmartManagement.onlyBuilder) {
      WidgetsBinding.instance.addPostFrameCallback((timeStamp) {
        YuroInstance().removeDependencyByRoute(reference);
      });

      MiddlewareRunner(middlewares).runPageDispose();
    }
  }

  @override
  bool canTransitionTo(TransitionRoute nextRoute) {
    // 全屏对话框不执行TransitionTo动画
    return nextRoute is PageRoute && !nextRoute.fullscreenDialog;
  }

  bool isPopGestureEnable<T>(PageRoute<T> route) {
    if (route.isFirst ||
        route.willHandlePopInternally ||
        // ignore: invalid_use_of_protected_member
        route.hasScopedWillPopCallback ||
        route.fullscreenDialog ||
        route.animation.status != AnimationStatus.completed ||
        route.secondaryAnimation.status != AnimationStatus.dismissed ||
        isPopGestureInProgress(route)) return false;

    return true;
  }

  bool isPopGestureInProgress<T>(PageRoute<T> route) {
    return route.navigator.userGestureInProgress;
  }

  _CupertinoBackGestureController<T> startPopGesture<T>(PageRoute<T> route) {
    assert(isPopGestureEnable(route));
    return _CupertinoBackGestureController<T>(controller: route.controller, navigator: route.navigator);
  }

  @override
  Widget buildTransitions(BuildContext context, Animation<double> animation, Animation<double> secondaryAnimation, Widget child) {
    final finalCurve = curve ?? Yuro.defaultTransitionCurve;
    final hasCurve = finalCurve != null;

    if (fullscreenDialog && transition == null) {
      return CupertinoFullscreenDialogTransition(
        primaryRouteAnimation: hasCurve ? CurvedAnimation(parent: animation, curve: finalCurve) : animation,
        secondaryRouteAnimation: secondaryAnimation,
        child: child,
        linearTransition: hasCurve,
      );
    }

    var popGestureChild = popGesture ?? Yuro.defaultPopGesture
        ? _CupertinoBackGestureDetector<T>(
            enableCallback: () => isPopGestureEnable<T>(this),
            onStartPopGesture: () => startPopGesture<T>(this),
            child: child,
          )
        : child;

    final temp = customTransition ?? Yuro.customTransition;
    if (temp != null) {
      return temp.buildTransitions(context, curve, alignment, animation, secondaryAnimation, popGestureChild);
    }
    final iosAnimation = animation;
    animation = CurvedAnimation(parent: animation, curve: finalCurve);
    switch (transition ?? Yuro.defaultTransitionCurve) {
      case Transition.FADE:
        return FadeUpwardsPageTransitionsBuilder().buildTransitions(this, context, animation, secondaryAnimation, child);
      case Transition.FADE_IN:
        return FadeInTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.LEFT_TO_RIGHT:
        return LeftToRightTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.RIGHT_TO_LEFT:
        return RightToLeftTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.UP_TO_DOWN:
        return UpToDownTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.DOWN_TO_UP:
        return UpToDownTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.LEFT_TO_RIGHT_WITH_FADE:
        return LeftToRightWithFadeTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.RIGHT_TO_LEFT_WITH_FADE:
        return RightToLeftWithFadeTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.ZOOM:
        return ZoomTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.TOP_LEVEL:
        return ZoomPageTransitionsBuilder().buildTransitions(this, context, animation, secondaryAnimation, child);
      case Transition.NO_TRANSITION:
        return NoTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.CUPERTINO:
        return CupertinoPageTransitionsBuilder().buildTransitions(this, context, iosAnimation, secondaryAnimation, child);
      // case Transition.CUPERTINO_DIALOG:
      case Transition.SIZE:
        return SizeTransitionBuilder().buildTransitions(context, curve, alignment, animation, secondaryAnimation, child);
      case Transition.NATIVE:
        return PageTransitionsTheme().buildTransitions(this, context, iosAnimation, secondaryAnimation, popGestureChild);
      default:
        return PageTransitionsTheme().buildTransitions(this, context, animation, secondaryAnimation, popGestureChild);
    }
  }
}

const double _kMinFlingVelocity = 1.0;
const double _kBackGestureWidth = 20.0;
const int _kMaxDroppedSwipePageForwardAnimationTime = 800;
// 滑动过程中释放手指,页面回到原始位置的时间
const int _kMaxPageBackAnimationTime = 300;

class _CupertinoBackGestureController<T> {
  _CupertinoBackGestureController({@required this.controller, @required this.navigator}) {
    assert(controller != null);
    assert(navigator != null);

    navigator.didStartUserGesture();
  }

  final AnimationController controller;
  final NavigatorState navigator;

  void dragUpdate(double delta) => controller.value = delta;

  void dragEnd(double velocity) {
    final animationCurve = Curves.fastLinearToSlowEaseIn;
    bool animationForward;

    if (velocity.abs() >= _kMinFlingVelocity) {
      animationForward = velocity <= 0;
    } else {
      animationForward = controller.value > 0.5;
    }

    if (animationForward) {
      final droppedPageForwardAnimationTime = min(
        lerpDouble(_kMaxDroppedSwipePageForwardAnimationTime, 0, controller.value).floor(),
        _kMaxPageBackAnimationTime,
      );
      controller.animateTo(1.0, duration: Duration(milliseconds: droppedPageForwardAnimationTime), curve: animationCurve);
    } else {
      navigator.pop();
      if (controller.isAnimating) {
        final droppedPageBackAnimationTime = lerpDouble(0, _kMaxDroppedSwipePageForwardAnimationTime, controller.value).floor();
        controller.animateTo(0.0, duration: Duration(milliseconds: droppedPageBackAnimationTime), curve: animationCurve);
      }
    }

    if (controller.isAnimating) {
      AnimationStatusListener animationStatusListener;
      animationStatusListener = (status) {
        navigator.didStopUserGesture();
        controller.removeStatusListener(animationStatusListener);
      };
      controller.addStatusListener(animationStatusListener);
    } else {
      navigator.didStopUserGesture();
    }
  }
}

class _CupertinoBackGestureDetector<T> extends StatefulWidget {
  final Widget child;
  final ValueGetter<bool> enableCallback;
  final ValueGetter<_CupertinoBackGestureController<T>> onStartPopGesture;

  const _CupertinoBackGestureDetector({
    Key key,
    @required this.enableCallback,
    @required this.onStartPopGesture,
    @required this.child,
  })  : assert(child != null),
        assert(enableCallback != null),
        assert(onStartPopGesture != null),
        super(key: key);

  @override
  _CupertinoBackGestureDetectorState<T> createState() => _CupertinoBackGestureDetectorState<T>();
}

class _CupertinoBackGestureDetectorState<T> extends State<_CupertinoBackGestureDetector<T>> {
  HorizontalDragGestureRecognizer _recognizer;
  _CupertinoBackGestureController<T> _backGestureController;

  @override
  void initState() {
    super.initState();
    _recognizer = HorizontalDragGestureRecognizer(debugOwner: this)
      ..onStart = _handlerOnStart
      ..onUpdate = _handlerOnUpdate
      ..onEnd = _handlerOnEnd
      ..onCancel = _handlerOnCancel;
  }

  @override
  void dispose() {
    _recognizer.dispose();
    super.dispose();
  }

  void _handlerOnStart(DragStartDetails details) {
    assert(mounted);
    assert(_backGestureController == null);
    _backGestureController = widget.onStartPopGesture();
  }

  void _handlerOnUpdate(DragUpdateDetails details) {
    assert(mounted);
    assert(_backGestureController != null);
    _backGestureController.dragUpdate(_convertToLogic(details.primaryDelta / context.size.width));
  }

  void _handlerOnEnd(DragEndDetails details) {
    assert(mounted);
    assert(_backGestureController != null);
    _backGestureController.dragEnd(_convertToLogic(details.velocity.pixelsPerSecond.dx / context.size.width));
    _backGestureController = null;
  }

  void _handlerOnCancel() {
    assert(mounted);
    _backGestureController?.dragEnd(0.0);
    _backGestureController = null;
  }

  double _convertToLogic(double value) {
    switch (Directionality.of(context)) {
      case TextDirection.rtl:
        return -value;
        break;
      default:
        return value;
        break;
    }
  }

  void _handlerOnPointerDown(PointerDownEvent event) {
    if (widget.enableCallback()) _recognizer.addPointer(event);
  }

  @override
  Widget build(BuildContext context) {
    assert(debugCheckHasDirectionality(context));
    var dragAreaWidth =
        Directionality.of(context) == TextDirection.ltr ? MediaQuery.of(context).padding.left : MediaQuery.of(context).padding.right;
    dragAreaWidth = max(dragAreaWidth, _kBackGestureWidth);

    return Stack(fit: StackFit.passthrough, children: [
      widget.child,
      PositionedDirectional(
          start: 0,
          top: 0,
          bottom: 0,
          width: dragAreaWidth,
          child: Listener(onPointerDown: _handlerOnPointerDown, behavior: HitTestBehavior.translucent))
    ]);
  }
}
